/**
 * SwarmScreen Project
 *
 * File:         ConnectionProfiler.java
 * RCS:          $Id: ConnectionProfiler.java,v 1.12 2009/04/16 17:59:17 drc915 Exp $
 * Description:  ConnectionProfiler class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Feb 21, 2009 at 10:21:50 AM
 * Language:     Java
 * Package:      edu.northwestern.swarmscreen.connection
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2009, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.swarmscreen.connection;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map.Entry;

import org.apache.commons.math.distribution.ChiSquaredDistributionImpl;
import org.apache.commons.math.stat.inference.ChiSquareTestImpl;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.peers.PeerManagerListener;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;

import edu.northwestern.swarmscreen.Main;
import edu.northwestern.swarmscreen.torrents.TorrentManager;
import edu.northwestern.swarmscreen.util.KSTest;

/**
 * @author David Choffnes
 *
 */
public class ConnectionProfiler implements PeerManagerListener, UTTimerEventPerformer {

	
	private static final boolean DROP_LONG_CONNECTIONS = false;

	private static final boolean DEBUG = false;

	private static ConnectionProfiler self;
	
	private HashMap<InetAddress, Integer> userConnections = 
		new HashMap<InetAddress, Integer>();
	
	private HashMap<InetAddress, Integer> transientConnections = 
		new HashMap<InetAddress, Integer>();
	
	private HashSet<Peer> addedPeers = new HashSet<Peer>();

	/** longest time that a peer can be connected*/
	private long MAX_CONNECTION_TIME = 15 *1000; // currently 2 minutes

	private int longestConnectionTime = -1;


	
//	ChiSquareTestImpl cs = new ChiSquareTestImpl();
//	ChiSquaredDistributionImpl csDist = new ChiSquaredDistributionImpl(1.0);

	private HashMap<Integer, Double> userRatio=null;
	private HashMap<Integer, Double> transientRatio=null;
	
	private HashMap<InetAddress, Integer> openUserConnections=null;
	private HashMap<InetAddress, Integer> openTransientConnections=null;

	private int transientConTime=-1;
	private int userConTime=-1;
	private double similarity = -1;
	private double ratio = -1;
	
	public static ConnectionProfiler getInstance(){
		if (self==null){
			self = new ConnectionProfiler();
//			self.csDist.setDegreesOfFreedom(1);
		}
		return self;
	}

	
	public void peerAdded(PeerManager manager, Peer peer) {
		synchronized(addedPeers){addedPeers.add(peer);}
	}

	
	public void peerRemoved(PeerManager manager, Peer peer) {
		synchronized (addedPeers) {
			if (addedPeers.remove(peer))
				// get connection time
				getPeerConnectionTime(peer);
		}
	

	}


	/**
	 * @param peer
	 */
	public void getPeerConnectionTime(Peer peer) {
		int time = (int)
		(peer.getStats().getTimeSinceConnectionEstablished()/1000);
		HashMap<InetAddress, Integer> cons;
		// figure out which map to put it in
		try {
			if (TorrentManager.getInstance().isUserDownload(peer.getManager().getDownload())){
				cons = userConnections;
			} else cons = transientConnections;
		} catch (DownloadException e) {		
			e.printStackTrace();
			return;
		}
		// put data in map
		InetAddress addr;
		try {
			addr = InetAddress.getByName(peer.getIp());	
			synchronized (cons) {
				if (cons.get(addr)==null) cons.put(addr, time);
				else cons.put(addr, time+cons.get(addr));
				if (cons.get(addr)>longestConnectionTime)
					longestConnectionTime = cons.get(addr);
			}
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
	}

	public void updateConnectionStats() {
		// can't seem to stop missing some...
		//addAllPeers();
		
		// this is a signal to update connection stats
		LinkedList<Peer> toRemove = new LinkedList<Peer>();
		
		// bucket the connection times, find distributions for each
		userRatio = new HashMap<Integer, Double>();
		transientRatio = new HashMap<Integer, Double>();
		userConTime = 0; transientConTime=0;
		synchronized (addedPeers){			
			// remove dead peers
			for (Peer p : addedPeers){
				try {p.getManager().getDownload();} catch (Exception e){ // remove all dead peers
					toRemove.add(p);
					//getPeerConnectionTime(p); // TODO figure out how to determine peer as transient or not...
				}
			}
			addedPeers.removeAll(toRemove);
			toRemove.clear();
				
			// skip this for now
			for (Peer p : addedPeers){
				if (DROP_LONG_CONNECTIONS && 
						p.getStats().getTimeSinceConnectionEstablished()>MAX_CONNECTION_TIME ){
					//p.getConnection().close(); // TODO does this trigger peer removed?
					toRemove.add(p);
				}
			}
			for (Peer p : toRemove) p.getManager().removePeer(p);
			
			// update stats for currently connected peers
			HashMap<Integer, Double> ratio;		
			HashMap<InetAddress, Integer> openConns;
			openUserConnections= new HashMap<InetAddress, Integer>();
			openTransientConnections=new HashMap<InetAddress, Integer>();

			//for (Download d : Main.getPluginInterface().getDownloadManager().getDownloads()){
			//	if (d.getPeerManager()==null) continue;		
			//	System.out.println("Dl: "+d.getName()+" Peers: "+d.getPeerManager().getPeers().length);
				for (Peer p : addedPeers){ //d.getPeerManager().getPeers()){					
					try {
						if (TorrentManager.getInstance().isUserDownload(p.getManager().getDownload())){
							ratio = userRatio;	
							openConns = openUserConnections;
						} else {
							ratio = transientRatio;
							openConns = openTransientConnections;
						}
						boolean user = ratio==userRatio;
						long connTime = p.getStats().getTimeSinceConnectionEstablished();//d.getPeerManager().createPeerStats(p).getTimeSinceConnectionEstablished();
						if (connTime==0) continue;
						try {
							openConns.put(InetAddress.getByName(p.getIp()), (int)connTime/1000);
						} catch (UnknownHostException e) {e.printStackTrace();}
						int bucket = (int) (connTime/MAX_CONNECTION_TIME);				
						if (ratio.get(bucket)==null) ratio.put(bucket, 1.0);
						else ratio.put(bucket, ratio.get(bucket)+1);
						if (user) userConTime+=connTime/1000;
						else transientConTime+=connTime/1000;
					} catch (DownloadException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}
		//}

		for (HashMap<InetAddress, Integer> cons : 
			new HashMap[]{userConnections, transientConnections}){
			synchronized (cons){
				boolean user = cons==userConnections;
				HashMap<Integer, Double> ratios = (user ? 
						userRatio : transientRatio);
				double size = cons.size();
				for (Entry<InetAddress, Integer> ent : cons.entrySet()){
					if (ent.getValue()==0) continue;
					int bucket = (int) (ent.getValue()/(MAX_CONNECTION_TIME/1000.0));				
					if (ratios.get(bucket)==null) ratios.put(bucket, 1.0);
					else ratios.put(bucket, ratios.get(bucket)+1);
					if (user) userConTime+=ent.getValue();
					else transientConTime+=ent.getValue();
				}

			} // end synchronized
		} // end for each connection set
		
		// put in ratios
		for (HashMap<Integer, Double> ratios : 
			new HashMap[]{userRatio, transientRatio}){
			
			for (int bucket : ratios.keySet()){		
				ratios.put(bucket, ratios.get(bucket));
			}
		}
		
		
		
		// now we have normalized ratios and total connection times
		ratio = -1;
		similarity = -1; // resets values so they are recomputed on demand
	}
	
	private void addAllPeers() {
		for (Download d: Main.getPluginInterface().getDownloadManager().getDownloads()){
			if (d.getPeerManager()!=null){
				synchronized (addedPeers){
					for (Peer p : d.getPeerManager().getPeers()){
						addedPeers.add(p);
					}
				}
			}
		}
		
	}


	/**
	 * Returns the total user connection time divided by 
	 * the total transient connection time. Should be kept 
	 * around 1.
	 * 
	 * @return
	 */
	public double getConnectionTimesRatio(){
		if (ratio < 0){
		if (userConTime <= 0 || transientConTime <= 0) return -1;
		 ratio  = ((double)transientConTime)/userConTime;
		} 
		return ratio;
	}
	
	/**
	 * Returns the cosine similarity of the two types of 
	 * downloads. 
	 * 
	 * @return
	 */
	public double getConnectionDistributionSimilarity(){
		if (!DEBUG) return -1;
		if (similarity <0){
			if (userRatio==null || transientRatio==null || 
					(userRatio.size()==0 && transientRatio.size()==0)) return -1;
			similarity = KSTestResult();//chiSquaredTestResult();
			/*HashMap<Integer, Double> userRatioNorm = 
				new HashMap<Integer, Double>();
			userRatioNorm.putAll(userRatio);
			HashMap<Integer, Double> transientRatioNorm = 
				new HashMap<Integer, Double>();
			transientRatioNorm.putAll(transientRatio);
			// normalize
			for (HashMap<Integer, Double> ratios : 
				new HashMap[]{userRatioNorm, transientRatioNorm}){
				double size =0.0;
				for (double value : ratios.values()){		
					size+=value;
				}
				for (int bucket : ratios.keySet()){		
					ratios.put(bucket, ratios.get(bucket));
				}
			}

			similarity  = Util.getCosineSimilarity(userRatioNorm, transientRatioNorm);*/
		} 
		return similarity;
	}
	
	/*public double chiSquaredTestResult(){
		if (userRatio==null || transientRatio==null || 
				(userRatio.size()==0 && transientRatio.size()==0)) return -2;
		
		// find size of the array
		int max = -1;
		for (int i : userRatio.keySet()) if (i>max) max = i;
		for (int i : transientRatio.keySet()) if (i>max) max = i;
		if (max < 1) return -2;
		//int max2 = Collections.max(userRatio.keySet());
		//if (Collections.max(transientRatio.keySet())<max2) max2 = Collections.max(transientRatio.keySet());
		
		// create and fill arrays
		long userRatioA[] = new long[max+1];
		long transientRatioA[] = new long[max+1];		
		for (Entry<Integer, Double> ent : userRatio.entrySet()){
			if (ent.getKey()<=max) userRatioA[ent.getKey()] = (long)ent.getValue().doubleValue();
		}
		for (Entry<Integer, Double> ent : transientRatio.entrySet()){
			if (ent.getKey()<=max) transientRatioA[ent.getKey()] = (long)ent.getValue().doubleValue();
		}		
		if (DEBUG) System.out.println(Arrays.toString(userRatioA)+"\n"+
				Arrays.toString(transientRatioA));
		
		// can't both be zero, find indexes to skip and compress arrays
		HashSet<Integer> toSkip = new HashSet<Integer>();
		for (int i = 0; i < max+1; i++){
			if (userRatioA[i]+transientRatioA[i]==0) toSkip.add(i);
		}
		if (toSkip.size()>0){
			for (long toCopy[] : new long[][]{userRatioA, transientRatioA}){
				long temp[] = new long[max+1-toSkip.size()];
				int i = 0, j =0;
				while (i < toCopy.length && j < temp.length){
					if (toSkip.contains(i)){
						i++;
						continue;
					}
					temp[j] = toCopy[i];
					j++;
					i++;
				}
				if (userRatioA==toCopy) userRatioA=temp;
				else transientRatioA=temp;
			}
		}
				
		double p = -2;
		try {
			//p = cs.chiSquareTestDataSetsComparison(userRatioA, transientRatioA);
			double dd[] = new double[userRatioA.length];
			for (int i = 0; i < dd.length; i++) {
				dd[i] = userRatioA[i];
				if (dd[i]==0) dd[i]+=0.01;
			}
			p = cs.chiSquareTestDataSetsComparison(userRatioA, transientRatioA);
		} catch (IllegalArgumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (MathException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return p;
	}*/
	
	public double KSTestResult(){
		
		
		ArrayList<Double> userCons = new ArrayList<Double>();
		ArrayList<Double> transCons = new ArrayList<Double>();
		for (Entry<InetAddress, Integer> ent : userConnections.entrySet()){
			double time = ent.getValue().doubleValue();
			if (openUserConnections.containsKey(ent.getKey())){
				time+=openUserConnections.get(ent.getKey());
			}
			if (time>0)userCons.add(time);
		}
		for (Entry<InetAddress, Integer> ent : openUserConnections.entrySet()){
			if (!userConnections.containsKey(ent.getKey())){
				if (ent.getValue()>0)userCons.add(ent.getValue().doubleValue());
			}
		}
		for (Entry<InetAddress, Integer> ent : transientConnections.entrySet()){
			double time = ent.getValue().doubleValue();
			if (openTransientConnections.containsKey(ent.getKey())){
				time+=openTransientConnections.get(ent.getKey());
			}
			if (time>0)transCons.add(time);
		}
		for (Entry<InetAddress, Integer> ent : openTransientConnections.entrySet()){
			if (!transientConnections.containsKey(ent.getKey())){
				if (ent.getValue()>0)transCons.add(ent.getValue().doubleValue());
			}
		}
		
		
		double p = -2;
		if (userCons.size()<3 || transCons.size()<3) return p;
		Collections.sort(userCons);
		Collections.sort(transCons);
		if (transCons.size()>userCons.size()){			
			double factor = ((double)transCons.size())/userCons.size() - 1;
			double f2 = 1+factor;
			double current = 0;
			for (int i = 0; i < transCons.size(); i++){
				current+=factor;
				if (current>=f2){
					transCons.remove(i);
					i--;
					current -=f2;
				}
			}
//			System.out.println("Trans: "+transCons.size()+"User: "+userCons.size());
		}
		
		if (DEBUG) System.out.println(userCons.toString());
		if (DEBUG) System.out.println(transCons.toString());
		double uCons[] = new double[userCons.size()];
		double tCons[] = new double[transCons.size()];
		for (int i = 0; i < userCons.size(); i++) uCons[i] = userCons.get(i);
		for (int i = 0; i < transCons.size(); i++) tCons[i] = transCons.get(i);
		p = KSTest.KolmogorovTest(uCons.length, uCons, tCons.length, tCons, "");
		return p;
	}



	public void perform(UTTimerEvent event) {
		
		updateConnectionStats();
		//if (DEBUG) System.out.println("Chi square: "+chiSquaredTestResult());
		if (DEBUG) System.out.println("KS test: "+KSTestResult());
		if (DEBUG) System.out.println("User: "+userConTime+" Trans: "+transientConTime);
		//TorrentManager.getInstance().fixTransferRates();
	}

}
