/**
 * SwarmScreen Project
 *
 * File:         TorrentManager.java
 * RCS:          $Id: TorrentManager.java,v 1.13 2009/04/16 17:59:17 drc915 Exp $
 * Description:  TorrentManager class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Feb 21, 2009 at 10:22:22 AM
 * Language:     Java
 * Package:      edu.northwestern.swarmscreen.torrents
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2009, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.swarmscreen.torrents;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.Map.Entry;

import org.gudy.azureus2.plugins.PluginConfig;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;
import org.gudy.azureus2.plugins.download.DownloadListener;
import org.gudy.azureus2.plugins.download.DownloadManagerListener;
import org.gudy.azureus2.plugins.download.DownloadPeerListener;
import org.gudy.azureus2.plugins.download.DownloadRemovalVetoException;
import org.gudy.azureus2.plugins.logging.LoggerChannel;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.Torrent;
import org.gudy.azureus2.plugins.torrent.TorrentException;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;
import org.gudy.azureus2.plugins.utils.resourcedownloader.ResourceDownloader;
import org.gudy.azureus2.plugins.utils.resourcedownloader.ResourceDownloaderException;
import org.gudy.azureus2.plugins.utils.resourcedownloader.ResourceDownloaderFactory;

import edu.northwestern.swarmscreen.Main;
import edu.northwestern.swarmscreen.SSConfiguration;
import edu.northwestern.swarmscreen.connection.ConnectionProfiler;
import edu.northwestern.swarmscreen.util.HashSetCache;
import edu.northwestern.swarmscreen.util.PluginInterface;
import edu.northwestern.swarmscreen.util.HashSetCache.RemovalListener;

/**
 * @author David Choffnes
 *
 */
public class TorrentManager implements DownloadManagerListener, DownloadListener, UTTimerEventPerformer {

	private static final boolean DEMO_ONLY = false;
	private org.gudy.azureus2.plugins.PluginInterface pi;
	private static TorrentManager self;
	
	public class DownloadStats {

		int numUserDls;
		int numTransientDls;
		int userDlRate;
		int transientDlRate;
		int userUlRate;
		int transientUlRate;
		int numActiveUserDls;
		
		
		public String toString() {
			// TODO Auto-generated method stub
			return numUserDls+"\t"+numTransientDls+"\t"+userDlRate+"\t"+
			transientDlRate+"\t"+userUlRate+"\t"+transientUlRate+"\t"+numActiveUserDls;
		}
		
	}
	
	public class TransientTorrentPool {
		
		HashSetCache<Download> pool;
		
		public TransientTorrentPool(){
			pool = new HashSetCache<Download>(maxTransientPoolSize);
			pool.addListener(new RemovalListener<Download>(){
				public void entryRemoved(Download removed) {
					try {
						removed.remove(true, true);
					} catch (DownloadException e) {e.printStackTrace();
					} catch (DownloadRemovalVetoException e) {e.printStackTrace();}
				}});
		}
		
		public synchronized Download activateDownload(){
			ArrayList<Download> pool2 = new ArrayList<Download>();
			pool2.addAll(pool);
			Collections.shuffle(pool2);
			Download toReturn = pool2.remove(pool2.size()-1);
			pool.remove(toReturn);
			return toReturn;
		}
		
		public synchronized void deactivateDownload(Download d){
			pool.add(d);
		}

		public synchronized boolean isEmpty() {			
			return pool.isEmpty();
		}

		public synchronized void deleteAll() {
			for (Download d : pool){
				d.stopDownload();
				try {
					d.remove(true, true);
				} catch (DownloadException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (DownloadRemovalVetoException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			
		}

		public synchronized void stopDownloads() {
			for (Download d : pool){
				d.stopDownload();
			}
			
		}
	}
	
	/** total slots for user and transient torrents */
	private int totalTorrentSlots = 10; // TODO set to max active torrents, TODO make param
	/** measure of how much overhead to allow: SPF/100 */
	private double performanceOverhead = 1;
	
	/** set of all active user downloads */
	HashSet<Download> userDls = new HashSet<Download>();
	/** set of all paused user downloads */
	HashSet<Download> userPausedDls = new HashSet<Download>();
	/** set of all active and paused user downloads */
	HashSet<Download> allUserDls = new HashSet<Download>();
	HashMap<Download, PeerManager> managers = new HashMap<Download, PeerManager>();
	
	ArrayList<Download> pendingDeletes = new ArrayList<Download>();
	
	HashSet<Download> transientDls = new HashSet<Download>();
	private HashSet<String> transientNames = new HashSet<String>();
	private long maxFileSize = SSConfiguration.getInstance().getMaxFileSize()*1024L*1024; // 2GB by default TODO make param
	
	private long lastRotateTime = -1;
	private boolean notEnabledWarningShown;
	private boolean noTorrentDirWarningShown;
	private boolean noUrlsWarningShown;
	private double supplementFactor = 0.1;
	private double lowWaterRatio = 0.9;
	private double highWaterRatio = 1.1;
	private int maxDownloadsDefault = -1;
	private int maxUploadsDefault = -1;
	private long userRotateInterval = 2*60*1000; // 3 minutes
	private long lastUserRotateTime = System.currentTimeMillis();
	private int maxTransientsPerInterval = 2;
	private boolean awaitingTransient = false;
	
	private TransientTorrentPool transientPool = new TransientTorrentPool();
	private static int maxTransientPoolSize = 10;

	public TorrentManager(org.gudy.azureus2.plugins.PluginInterface pi_az) {
		this.pi = pi_az;
		self = this;
		for (Download d : pi.getDownloadManager().getDownloads()){
			if (isDownloadActive(d)){
				userDls.add(d);
				checkAddPeerListener(d);
			}
			d.addListener(this);
		}
		try {
			maxDownloadsDefault = pi.getPluginconfig().getUnsafeIntParameter(PluginConfig.CORE_PARAM_INT_MAX_DOWNLOADS);
			maxUploadsDefault = pi.getPluginconfig().getUnsafeIntParameter(PluginConfig.CORE_PARAM_INT_MAX_UPLOADS);
		
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	/**
	 * @param d
	 */
	private void checkAddPeerListener(Download d) {
		if (d.getPeerManager()==null) {
			managers.put(d, null);
		}
		else {
			d.getPeerManager().addListener(ConnectionProfiler.getInstance());
			managers.put(d, d.getPeerManager());
		}
	}
	
	public static TorrentManager getInstance(){
		return self;
	}

	private boolean isDownloadActive(Download d) {
		return d.getState()<=Download.ST_SEEDING;
	}

	
	public void downloadAdded(Download download) {
		if (isDownloadActive(download) && !isTransientTorrent(download)){
			synchronized(userDls){
				userDls.add(download);
				allUserDls.add(download);
			}			
		} 
		download.addListener(this);
		checkAddPeerListener(download);
		
	}

	private boolean isTransientTorrent(Download download) {
		synchronized(transientDls){
		if (transientDls.contains(download) || 
				transientNames.contains(download.getTorrent().getName()) ) 
			return true;
		}
		return false;
	}

	
	public void downloadRemoved(Download download) {
		synchronized(userDls){
			userDls.remove(download);			
			allUserDls.remove(download);
		}	
		managers.remove(download);
	}

	
	public void positionChanged(Download download, int oldPosition,
			int newPosition) {
		// don't care
		
	}

	
	public void stateChanged(Download download, int old_state, int new_state) {		
		if (new_state <=Download.ST_SEEDING) {
			if (isTransientTorrent(download)){
				if (!transientDls.contains(download))
					synchronized(transientDls){transientDls.add(download);}
				if (new_state==Download.ST_SEEDING){
					removeTransientDownload(download);
				}
			}
			else{
				userDls.add(download);
				userPausedDls.remove(download);
				allUserDls.add(download);
				if (allUserDls.size()>0 && transientDls.size()==0){
					if (canRunSS()){
					stopUserDownload(download, true);
					awaitingTransient  = true;
					}
				}
			}
			checkAddPeerListener(download);
		} else {
			managers.remove(download.getPeerManager());
		}
//		else {
//			if (isTransientTorrent(download))
//				synchronized(transientDls){transientDls.remove(download);}
//				else userDls.remove(download);
//		}
		
	}

	
	private void removeTransientDownload(Download d) {
//		if (d.getPeerManager()!=null){
//			for (Peer p : d.getPeerManager().getPeers()){
//				ConnectionProfiler.getInstance().getPeerConnectionTime(p);
//			}
//		}
		
		final Download d2 = d;
		if (d.getState()==Download.ST_SEEDING){
			transientDls.remove(d);
			d.stopDownload();
			try {
				d.remove(true, true);
			} catch (DownloadException e) {	e.printStackTrace();} 
			catch (DownloadRemovalVetoException e) {e.printStackTrace();}	
			
			
			transientNames.remove(d.getTorrent().getName());
			
			// TODO fetch new torrent
		}
		else {
			transientDls.remove(d);
			if (ConnectionProfiler.getInstance().getConnectionTimesRatio()>highWaterRatio) {
				d.stopDownload();
			} else {
				d.setDownloadRateLimitBytesPerSecond(-1);
				d.setUploadRateLimitBytesPerSecond(-1);
				if (d.getPeerManager()!=null){
					for (Peer p : d.getPeerManager().getPeers()){
						d.getPeerManager().removePeer(p);
					}
				}
			}
			transientNames.remove(d.getTorrent().getName());
			transientPool.deactivateDownload(d);
			//synchronized (pendingDeletes){pendingDeletes.add(d);}
		}
		
	}

	public void perform(UTTimerEvent event) {
		// periodically do the following:
		// 1) rotate torrents
		// 2) change connection rates
		maxFileSize = SSConfiguration.getInstance().getMaxFileSize()*1024L*1024;
		if (!canRunSS()) return;
		
		// if first time, load up transient torrents
		if (transientPool.isEmpty()){
			loadTransientTorrents();
		}
		
		// weirdness w/ losing peer managers...
		synchronized (transientDls){
			//for (HashSet<Download> dls : new HashSet[]{userDls, transientDls}){
//				for (Download d : pi.getDownloadManager().getDownloads()){
//					if (d.getPeerManager()!=null && isDownloadActive(d)){
//						d.getPeerManager().removeListener(ConnectionProfiler.getInstance());
//						d.getPeerManager().addListener(ConnectionProfiler.getInstance());
//					}
//				}
			//}
			
			for (Download d : transientDls){
				if (d.getState()>Download.ST_SEEDING ){
					d.startDownload(true);
				}
			}
		}
		
		performanceOverhead = SSConfiguration.getInstance().getPrivacyLevel()/100.0;
		
		// add any missing peer manager listeners
		LinkedList<Download> dls = new LinkedList<Download>();
		for (Entry<Download, PeerManager> ent : managers.entrySet()){
			if (ent.getValue()==null){
				if (ent.getKey().getPeerManager()!=null){
					ent.getKey().getPeerManager().addListener(ConnectionProfiler.getInstance());
					dls.add(ent.getKey());
				}
			}
		}
		for (Download d : dls) managers.put(d, d.getPeerManager());
		
		// has its own timer now
		//ConnectionProfiler.getInstance().updateConnectionStats();
		
		Set<URL> torrents = TorrentFetcher.getInstance().getRandomTorrents();
		
		rotateTorrents(torrents);
		
		//fixTransferRates();
		
		// TODO adjust the number of peers/seeds for each download 
		// to make them the same as the real downloads
		
	}

	public void loadTransientTorrents() {
		if (!canRunSS()) return;
		ArrayList<URL> urls = new ArrayList<URL>();
		urls.addAll(TorrentFetcher.getInstance().getRandomTorrents());
		Collections.shuffle(urls);
		
		for (int i = 0; i < maxTransientPoolSize  && i < urls.size(); i++){
			URL transientURL = urls.get(i);

			
			ResourceDownloaderFactory factory = pi.getUtilities().getResourceDownloaderFactory();
			ResourceDownloader rd = factory.create(transientURL);				
			ResourceDownloader rdt = factory.getTimeoutDownloader(rd, 15*1000);
			try {
				final Torrent t = pi.getTorrentManager().createFromBEncodedInputStream(rdt.download());	
				t.setDefaultEncoding();
				
				if (t == null || t.getSize()>maxFileSize || userDownloadExists(t)){					
					urls.remove(i);
					i--;
					continue;
				}		
				
				transientNames.add(t.getName());
			Download d = pi.getDownloadManager().addDownload(t, 
					null, 
					new File(SSConfiguration.getInstance().getTorrentDataDir()));
			d.stopDownload();			
			transientPool.deactivateDownload(d);
			transientNames.remove(t.getName());
			d.addPeerListener(new DownloadPeerListener(){
				public void peerManagerAdded(Download download,
						PeerManager peer_manager) {
					checkAddPeerListener(download);
				}
				
				public void peerManagerRemoved(
						Download download,
						PeerManager peer_manager) {
					peer_manager.removeListener(ConnectionProfiler.getInstance());
					managers.remove(download);
				}});

		} catch (DownloadException e) {
			e.printStackTrace();
		} catch (TorrentException e) {
			i--;
			continue;
		} catch (ResourceDownloaderException e) {
			i--;
			continue;
		}
		
		}
		
	}

	public void fixTransferRates() {
		
		
			int uploadRateUser = getUploadRate(userDls);
			int uploadRateTransient = getUploadRate(transientDls);
			int dlRateUser = getDownloadRate(userDls);
			int dlRateTransient = getDownloadRate(transientDls);
		
			if (awaitingTransient && (dlRateTransient>0 || uploadRateTransient>0 || 
					ConnectionProfiler.getInstance().getConnectionTimesRatio()> highWaterRatio)){
				resetTimers();
				activateUserDls();
			} else if (!awaitingTransient 
					&& !(dlRateTransient>0 || uploadRateTransient>0)
					&& ConnectionProfiler.getInstance().getConnectionTimesRatio()<lowWaterRatio){
				awaitingTransient = true;
				for (Download d : allUserDls) stopUserDownload(d, false);
			}
			
			synchronized(transientDls){
			int totalActiveTransients = 0;
			for (Download dl : transientDls) if (dl.getState()==Download.ST_DOWNLOADING||dl.getState()==Download.ST_SEEDING)
				totalActiveTransients++;
		
		
			if (uploadRateUser>=0 && uploadRateTransient>0) {
				int upSum = uploadRateUser + uploadRateTransient;
				//double ratio = ((double)uploadRateTransient)/uploadRateUser;
				//if (ratio > performanceOverhead){
					double totalForTransient = upSum/(1+(1/performanceOverhead));
					double perTransientRate = totalForTransient/totalActiveTransients;
					if (uploadRateUser==0) perTransientRate =0;
					double extraBw = 0;
					ArrayList<Download> toAdd = new ArrayList<Download>();
					for (Download d : transientDls){
						long temp = (int)perTransientRate -d.getStats().getUploadAverage();//d.getUploadRateLimitBytesPerSecond()-d.getStats().getUploadAverage();						
						if (temp>1000) extraBw+=temp;
						else toAdd.add(d);
						d.setUploadRateLimitBytesPerSecond((int)perTransientRate);
					}
					double extraPerTorrent = extraBw/toAdd.size();
					for (Download d : toAdd){
						d.setUploadRateLimitBytesPerSecond(
								(int)(d.getUploadRateLimitBytesPerSecond()+extraPerTorrent));
					}
				//}
			}
		
		
			if (dlRateUser>=0 && dlRateTransient>0) {
				int upSum = dlRateUser + dlRateTransient;
				//double ratio = ((double)dlRateTransient)/dlRateUser;
				//if (ratio > performanceOverhead){
					double totalForTransient = upSum/(1+(1/performanceOverhead));
					double perTransientRate = totalForTransient/totalActiveTransients;
					if (dlRateUser==0) perTransientRate =0;
					double extraBw = 0;
					ArrayList<Download> toAdd = new ArrayList<Download>();
					for (Download d : transientDls){
						long temp = (int)perTransientRate-d.getStats().getDownloadAverage();//d.getDownloadRateLimitBytesPerSecond()-d.getStats().getDownloadAverage();						
						if (temp>1000) extraBw+=temp;
						else toAdd.add(d);
						d.setDownloadRateLimitBytesPerSecond((int)perTransientRate);
					}
					double extraPerTorrent = extraBw/toAdd.size();
					for (Download d : toAdd){
						d.setDownloadRateLimitBytesPerSecond(
								(int)(d.getDownloadRateLimitBytesPerSecond()+extraPerTorrent));
					}
				//}
			}
		} // end sync transient dls
		
	}

	private boolean canRunSS() {
		if (DEMO_ONLY) return false;
		// checks for all the contingencies:
		// 1) Directory set for downloads
		// 2) Explicit enabling of the plugin
		// 3) URLs set for torrents
		boolean canRun = true;
		String messageId = "";
		if (!SSConfiguration.getInstance().isEnabled()){
			if (notEnabledWarningShown) return false;
			messageId = "ss.mustEnable";
			notEnabledWarningShown = true;
			canRun = false;
		} else if (SSConfiguration.getInstance().getTorrentDataDir()==null ||
				SSConfiguration.getInstance().getTorrentDataDir().equals("null") || 
				SSConfiguration.getInstance().getTorrentDataDir().length()==0){
			if (noTorrentDirWarningShown) return false;
			messageId = "ss.noTorrentDirectory";
			noTorrentDirWarningShown = true;
			canRun = false;
		} else if (SSConfiguration.getInstance().getTorrentURLs()==null ||
				SSConfiguration.getInstance().getTorrentURLs().length==0){
			if (noUrlsWarningShown) return false;
			messageId = "ss.mustEnable";
			noUrlsWarningShown = true;
			canRun = false;
		}

		if (!canRun) Main.getLogger().logAlert(LoggerChannel.LT_WARNING, 
				pi.getUtilities().getLocaleUtilities().
				getLocalisedMessageText(messageId));
		return canRun;
	}

	private void rotateTorrents(Set<URL> torrents) {
		if (Main.isShuttingDown()) return;
		if (hasStoppingTorrent()){
			resetTimers();
		}
		
		//ArrayList<URL> transients = new ArrayList<URL>();
		//transients.addAll(torrents);
		int rotateIntervalFactorTransient = 1;
		int rotateIntervalFactorUser = 1;
		
		int slotsForTransientTorrents = (int) Math.ceil(allUserDls.size()*performanceOverhead);
		int slotsForRealTorrents = totalTorrentSlots-slotsForTransientTorrents;		
		
		// add to this the number of torrents to account for imbalanced 
		// connection behavior
		double connectionRatio = ConnectionProfiler.getInstance().getConnectionTimesRatio();
		if (connectionRatio>0 && connectionRatio < lowWaterRatio && performanceOverhead>0 ){
			double tempRatio = getCurrentConnectionRatio();
			if (tempRatio<connectionRatio) connectionRatio = tempRatio;
			int toAdd = (int)((1-getCurrentConnectionRatio())/supplementFactor);
			slotsForTransientTorrents += toAdd;
			rotateIntervalFactorTransient = toAdd;
			
			// update positions
			int i = 0;
			synchronized (transientDls){
				
				for (Download d : transientDls){
					d.setPosition(i+1);
					i++;
				}
			}
			synchronized (allUserDls){
				for (Download d : allUserDls){
					d.setPosition(i+1);
					i++;
				}
			}
			//lastRotateTime=-1;			
		} else if (connectionRatio>0 && connectionRatio > highWaterRatio){
			transientPool.stopDownloads();
			int toRemove = (int) ((connectionRatio-1)/supplementFactor);
			slotsForTransientTorrents-=toRemove;
			if (slotsForTransientTorrents<1) slotsForTransientTorrents=1;
			rotateIntervalFactorUser = toRemove;
			lastRotateTime=-1;
			
			// update positions
			int i = 0;
			synchronized (allUserDls){
				for (Download d : allUserDls){
					d.setPosition(i++);
				}
			}
			synchronized (transientDls){				
				for (Download d : transientDls){
					d.setPosition(i+1);
					i++;
				}
			}
		}
		if (slotsForTransientTorrents < 0) slotsForTransientTorrents = 0;
		// if fewer slots than active torrents, rotate them
		// TODO rotate regardless of slots
		if (allUserDls.size()>0 && performanceOverhead>0 && 
				(slotsForRealTorrents < userDls.size()+userPausedDls.size() 
				  || System.currentTimeMillis()-lastUserRotateTime > 
				  (userRotateInterval*rotateIntervalFactorUser)/performanceOverhead) ){
			lastUserRotateTime = System.currentTimeMillis();
			ArrayList<Download> temp = new ArrayList<Download>();
			Download d = null;
			if (userDls.size()>0){
				synchronized(userDls){
					temp.addAll(userDls);
					d = temp.get((int)(Math.random()*userDls.size()));
	//				if (d.getPeerManager()!=null){
	//					for (Peer p : d.getPeerManager().getPeers()){
	//						ConnectionProfiler.getInstance().getPeerConnectionTime(p);
	//					}
	//				}	
					userDls.remove(d);	
				}
				stopUserDownload(d, false);
			}
			
			Download d2;
			temp.clear();
			if (userPausedDls.size()>0){
				synchronized(userPausedDls){
					temp.addAll(userPausedDls);				
					d2 = temp.get((int)(Math.random()*userPausedDls.size()));
					userPausedDls.remove(d2);
					userDls.add(d2);
					startUserDownload(d2);					
				}			
			}
			if (d!=null)synchronized(userPausedDls){userPausedDls.add(d);}
		}
		
		// now remove any that have no connections
		LinkedList<Download> dls = new LinkedList<Download>();
		synchronized (transientDls){				
			dls.addAll(transientDls);
		}
		
		for (Download d : dls){
			if (d.getPeerManager()!=null && 
					d.getStats().getTimeStarted()<System.currentTimeMillis()-2*60*1000
					&& pi.getDownloadManager().getStats().getDataReceiveRate()>0){
				if (d.getPeerManager().getStats().getConnectedLeechers()+
						d.getPeerManager().getStats().getConnectedSeeds()==0
						&& d.getDownloadRateLimitBytesPerSecond()<1000){
					removeTransientDownload(d);
				}
			}
		}
		
		// remove any existing torrents
		synchronized (transientDls){
		ArrayList<Download> dlsToRemove = new ArrayList<Download>();
		if (transientDls.size()>= slotsForTransientTorrents-1 && 
				transientDls.size()>0 &&
				pi.getUtilities().getCurrentSystemTime()-lastRotateTime>
				SSConfiguration.getInstance().getTorrentRotateInterval()*
				rotateIntervalFactorTransient){
			// at least remove one
			lastRotateTime=pi.getUtilities().getCurrentSystemTime();
			int numToRemove = transientDls.size()-slotsForTransientTorrents+1;
			while (numToRemove>0 && transientDls.size()>0){			
			ArrayList<Download> temp = new ArrayList<Download>();
			
				temp.addAll(transientDls);
				Collections.shuffle(temp);
				Download d = temp.get(0);
				removeTransientDownload(d);
				
			
			numToRemove--;
			}
			
		}
			
		}
	
		
		int numTorrents = slotsForTransientTorrents - transientDls.size();// + pendingDeletes.size();
		if (numTorrents > maxTransientsPerInterval ) numTorrents = maxTransientsPerInterval;
		// start up -- add extra torrents
		if (awaitingTransient) numTorrents*=2;
		System.out.println("Adding " + numTorrents + " torrents!");
		//Collections.shuffle(transients);
		while (numTorrents > 0 && !transientPool.isEmpty() //transients.size()>0
				&& (awaitingTransient || pi.getDownloadManager().getStats().getDataReceiveRate()>0)){			
//			int index = (int)(Math.random()*transients.size());
			
			
			Download d = transientPool.activateDownload();
			startTransientDownload(d);
			
			/*URL transientURL = transients.get(0);		
			
			ResourceDownloaderFactory factory = pi.getUtilities().getResourceDownloaderFactory();
			ResourceDownloader rd = factory.create(transientURL);				
			ResourceDownloader rdt = factory.getTimeoutDownloader(rd, 15*1000);
			try {
				final Torrent t = pi.getTorrentManager().createFromBEncodedInputStream(rdt.download());	
				t.setDefaultEncoding();
				
				if (t == null || t.getSize()>maxFileSize || userDownloadExists(t)){
					transients.remove(0);
					continue;
				}
				
				if (!Main.isShuttingDown())
				Main.getInstance().getTimer("TorrentManager").addEvent(System.currentTimeMillis(), new UTTimerEventPerformer(){

					
					public void perform(UTTimerEvent event) {
						try {				

							Download d = pi.getDownloadManager().addDownload(t, 
									null, 
									new File(SSConfiguration.getInstance().getTorrentDataDir()));
							
							d.addPeerListener(new DownloadPeerListener(){
								public void peerManagerAdded(Download download,
										PeerManager peer_manager) {
									checkAddPeerListener(download);
									
									
//									peer_manager.addListener(ConnectionProfiler.getInstance());
//									managers.put(download, download.getPeerManager());
								
									synchronized (pendingDeletes){
										for (Download d2 : pendingDeletes){												
											d2.stopDownload();
												try {
													d2.remove(true, true);
												} catch (DownloadException e) {	e.printStackTrace();} 
												catch (DownloadRemovalVetoException e) {e.printStackTrace();}	
												
												transientDls.remove(d2);
												transientNames.remove(d2.getTorrent().getName());
												
										}
										pendingDeletes.clear();
									}
								}
								
								public void peerManagerRemoved(
										Download download,
										PeerManager peer_manager) {
									peer_manager.removeListener(ConnectionProfiler.getInstance());
									managers.remove(download);
								}});
							
							synchronized(transientDls) {
								transientDls.add(d);
							}
							
							//d.setPosition(Integer.MAX_VALUE-transients.size());
						} catch (DownloadException e) {
							e.printStackTrace();
						}
						
					}});
			} catch (ResourceDownloaderException e) {
				e.printStackTrace();
				transients.remove(0);
				continue;
			} catch (TorrentException e) {
				e.printStackTrace();
			}
			transients.remove(0);*/
			
			
			numTorrents--;
			
		}
		
	}


	private void startTransientDownload(Download d) {
		transientNames.add(d.getTorrent().getName());
		transientDls.add(d);
		try {
			if (d.getState()==Download.ST_STOPPED) d.restart();
		} catch (DownloadException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		d.setDownloadRateLimitBytesPerSecond(0);
		d.setUploadRateLimitBytesPerSecond(0);
	}

	private void resetTimers() {
		lastRotateTime = System.currentTimeMillis();
		lastUserRotateTime = System.currentTimeMillis();
		
	}

	private boolean hasStoppingTorrent() {
		for (Download d : pi.getDownloadManager().getDownloads()){
			if (d.getState()==Download.ST_STOPPING) return true;
		}
		return false;
	}

	private double getCurrentConnectionRatio() {
		double userCons = 0;
		synchronized (allUserDls){
			for (Download d : allUserDls){
				if (d.getPeerManager()!=null){
					userCons+= d.getPeerManager().getPeers().length;
				}
			}
		}
		double transientCons = 0;
		synchronized (transientDls){
			for (Download d : transientDls){
				if (d.getPeerManager()!=null){
					transientCons+= d.getPeerManager().getPeers().length;
				}
			}
		}
		return transientCons/userCons;
	}

	private void stopUserDownload(Download d, boolean actuallyStop) {
		if (actuallyStop) d.stopDownload();
		else {
			d.setDownloadRateLimitBytesPerSecond(-1);
			d.setUploadRateLimitBytesPerSecond(-1);
			if (d.getPeerManager()!=null){
				for (Peer p : d.getPeerManager().getPeers()){
					d.getPeerManager().removePeer(p);
				}
			}
		}
	}
	
	private void startUserDownload(Download d){
		if (d.getState()==Download.ST_STOPPED){
			try {
				d.restart();
			} catch (DownloadException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		d.setDownloadRateLimitBytesPerSecond(0);
		d.setUploadRateLimitBytesPerSecond(0);
	}

	private boolean userDownloadExists(Torrent t) {
		synchronized (userDls){
			for (Download d : userDls){
				if (Arrays.equals(d.getTorrent().getHash(), t.getHash())){
					return true;
				}
			}
		}
		return false;
	}

	private int getUploadRate(HashSet<Download> dls) {
		int toReturn = 0;
		synchronized (dls){
			for (Download d : dls){
			toReturn+=d.getStats().getUploadAverage();
		}
		}
		return toReturn;
	}
	
	private int getDownloadRate(HashSet<Download> dls){
		int toReturn = 0;
		synchronized (dls){
			for (Download d : dls){
				if (d.getState()!=Download.ST_DOWNLOADING) continue;
				toReturn+=d.getStats().getDownloadAverage();
			}
		}
		return toReturn;
	}

	public boolean isUserDownload(Download download) {
		synchronized(userDls){
			return 
			allUserDls.contains(download) || 
			userDls.contains(download) || 
			userPausedDls.contains(download);
		}
	}


	public void deleteTransientTorrents(){
		ArrayList<Download> toDelete = new ArrayList<Download>();
		synchronized(transientDls){	toDelete.addAll(transientDls); }
		
			for (Download d : toDelete){
				try {
					if (d.getPeerManager()!=null){
						for (Peer p : d.getPeerManager().getPeers()){
							ConnectionProfiler.getInstance().getPeerConnectionTime(p);
						}
					}
					d.removeListener(this);
					d.stopDownload();
					d.remove(true, true);
				} catch (DownloadException e) {
					e.printStackTrace();
				} catch (DownloadRemovalVetoException e) {
					e.printStackTrace();
				}
			}
			
		transientPool.deleteAll();
		
		for (Download d : allUserDls){
			startUserDownload(d);
		}
	}

	public DownloadStats getDownloadStats() {
		DownloadStats ds = new DownloadStats();
		
		ds.userUlRate = getUploadRate(userDls);
		ds.transientUlRate = getUploadRate(transientDls);
		ds.userDlRate = getDownloadRate(userDls);
		ds.transientDlRate = getDownloadRate(transientDls);
		ds.numTransientDls = 0;
		for (Download dl : transientDls) if (dl.getState()==Download.ST_DOWNLOADING||dl.getState()==Download.ST_SEEDING)
			ds.numTransientDls++;
		ds.numUserDls = userDls.size();
		ds.numActiveUserDls = userDls.size();		
		return ds;
	}
	
	private void activateUserDls() {
		awaitingTransient = false;
		synchronized(allUserDls){
			for (Download d : allUserDls){
					startUserDownload(d);
				
			}
		}
		
	}

	public void transientPeerAdded() {
		if (awaitingTransient){
			activateUserDls();
		}
		
	}

}
