/**
 * SwarmScreen Project
 *
 * File:         KSTest.java
 * RCS:          $Id: KSTest.java,v 1.1 2009/03/06 14:40:14 drc915 Exp $
 * Description:  KSTest class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Mar 5, 2009 at 7:53:47 PM
 * Language:     Java
 * Package:      edu.northwestern.swarmscreen.util
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2009, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.swarmscreen.util;

import java.util.Arrays;

/**
 * @author David Choffnes
 *
 */
public class KSTest {
	//______________________________________________________________________________
	public static double KolmogorovTest(int na, double a[], int nb, double b[], String option)
	{
	//  Statistical test whether two one-dimensional sets of points are compatible
	//  with coming from the same parent distribution, using the Kolmogorov test.
	//  That is, it is used to compare two experimental distributions of unbinned data.
	//
	//  Input:
	//  a,b: One-dimensional arrays of length na, nb, respectively.
//	       The elements of a and b must be given in ascending order.
	//  option is a character string to specify options
//	         "D" Put out a line of "Debug" printout
//	         "M" Return the Maximum Kolmogorov distance instead of prob
	//
	//  Output:
	// The returned value prob is a calculated confidence level which gives a
	// statistical test for compatibility of a and b.
	// Values of prob close to zero are taken as indicating a small probability
	// of compatibility. For two point sets drawn randomly from the same parent
	// distribution, the value of prob should be uniformly distributed between
	// zero and one.
	//   in case of error the function return -1
	//   If the 2 sets have a different number of points, the minimum of
	//   the two sets is used.
	//
	// Method:
	// The Kolmogorov test is used. The test statistic is the maximum deviation
	// between the two integrated distribution functions, multiplied by the
	// normalizing factor (rdmax*sqrt(na*nb/(na+nb)).
	//
	//  Code adapted by Rene Brun from CERNLIB routine TKOLMO (Fred James)
	//   (W.T. Eadie, D. Drijard, F.E. James, M. Roos and B. Sadoulet,
//	      Statistical Methods in Experimental Physics, (North-Holland,
//	      Amsterdam 1971) 269-271)
	//
	//  Method Improvement by Jason A Detwiler (JADetwiler@lbl.gov)
	//  -----------------------------------------------------------
	//   The nuts-and-bolts of the TMath::KolmogorovTest() algorithm is a for-loop
	//   over the two sorted arrays a and b representing empirical distribution
	//   functions. The for-loop handles 3 cases: when the next points to be
	//   evaluated satisfy a>b, a<b, or a=b:
	//
//	      for (int i=0;i<na+nb;i++) {
//	         if (a[ia-1] < b[ib-1]) {
//	            rdiff -= sa;
//	            ia++;
//	            if (ia > na) {ok = true; break;}
//	         } else if (a[ia-1] > b[ib-1]) {
//	            rdiff += sb;
//	            ib++;
//	            if (ib > nb) {ok = true; break;}
//	         } else {
//	            rdiff += sb - sa;
//	            ia++;
//	            ib++;
//	            if (ia > na) {ok = true; break;}
//	            if (ib > nb) {ok = true; break;}
//	        }
//	         rdmax = Math.max(rdmax,Math.abs(rdiff));
//	      }
	//
	//   For the last case, a=b, the algorithm advances each array by one index in an
	//   attempt to move through the equality. However, this is incorrect when one or
	//   the other of a or b (or both) have a repeated value, call it x. For the KS
	//   statistic to be computed properly, rdiff needs to be calculated after all of
	//   the a and b at x have been tallied (this is due to the definition of the
	//   empirical distribution function; another way to convince yourself that the
	//   old CERNLIB method is wrong is that it implies that the function defined as the
	//   difference between a and b is multi-valued at x -- besides being ugly, this
	//   would invalidate Kolmogorov's theorem).
	//
	//   The solution is to just add while-loops into the equality-case handling to
	//   perform the tally:
	//
//	         } else {
//	            double x = a[ia-1];
//	            while(a[ia-1] == x && ia <= na) {
//	              rdiff -= sa;
//	              ia++;
//	            }
//	            while(b[ib-1] == x && ib <= nb) {
//	              rdiff += sb;
//	              ib++;
//	            }
//	            if (ia > na) {ok = true; break;}
//	            if (ib > nb) {ok = true; break;}
//	         }
	//
	//  NOTE1
	//  A good description of the Kolmogorov test can be seen at:
//	    http://www.itl.nist.gov/div898/handbook/eda/section3/eda35g.htm

	   String opt = option;
	   opt = opt.toUpperCase();

	   double prob = -1;
//	      Require at least two points in each graph
	   if (a==null || b==null || na <= 2 || nb <= 2) {
	      System.err.println("KolmogorovTest::Sets must have more than 2 points");
	      return prob;
	   }
//	     finalants needed
	   double rna = na;
	   double rnb = nb;
	   double sa  = 1./rna;
	   double sb  = 1./rnb;
	   double rdiff;
	   int ia,ib;
//	     Starting values for main loop
	   if (a[0] < b[0]) {
	      rdiff = -sa;
	      ia = 2;
	      ib = 1;
	   } else {
	      rdiff = sb;
	      ib = 2;
	      ia = 1;
	   }
	   double rdmax = Math.abs(rdiff);

//	    Main loop over point sets to find max distance
//	    rdiff is the running difference, and rdmax the max.
	   boolean ok = false;
	   for (int i=0;i<na+nb;i++) {
	      if (a[ia-1] < b[ib-1]) {
	         rdiff -= sa;
	         ia++;
	         if (ia > na) {ok = true; break;}
	      } else if (a[ia-1] > b[ib-1]) {
	         rdiff += sb;
	         ib++;
	         if (ib > nb) {ok = true; break;}
	      } else {
	         double x = a[ia-1];
	         while(a[ia-1] == x && ia <= na) {
	            rdiff -= sa;
	            ia++;
	         }
	         while(b[ib-1] == x && ib <= nb) {
	            rdiff += sb;
	            ib++;
	         }
	         if (ia > na) {ok = true; break;}
	         if (ib > nb) {ok = true; break;}
	      }
	      rdmax = Math.max(rdmax,Math.abs(rdiff));
	   }
//	    Should never terminate this loop with ok = kFALSE!

	   if (ok) {
	      rdmax = Math.max(rdmax,Math.abs(rdiff));
	      double z = rdmax * Math.sqrt(rna*rnb/(rna+rnb));
	      prob = KolmogorovProb(z);
	   }
	      // debug printout
	   if (opt.contains("D")) {
	     System.out.printf(" Kolmogorov Probability = %g, Max Dist = %g\n",prob,rdmax);
	   }
	   if(opt.contains("M")) return rdmax;
	   else                  return prob;
	}

	//______________________________________________________________________________
	public static double KolmogorovProb(double z)
	{
	   // Calculates the Kolmogorov distribution function,
	   //Begin_Html
	   /*
	   <img src="gif/kolmogorov.gif">
	   */
	   //End_Html
	   // which gives the probability that Kolmogorov's test statistic will exceed
	   // the value z assuming the null hypothesis. This gives a very powerful
	   // test for comparing two one-dimensional distributions.
	   // see, for example, Eadie et al, "statistocal Methods in Experimental
	   // Physics', pp 269-270).
	   //
	   // This function returns the confidence level for the null hypothesis, where:
	   //   z = dn*sqrt(n), and
	   //   dn  is the maximum deviation between a hypothetical distribution
	   //       function and an experimental distribution with
	   //   n    events
	   //
	   // NOTE: To compare two experimental distributions with m and n events,
	   //       use z = sqrt(m*n/(m+n))*dn
	   //
	   // Accuracy: The function is far too accurate for any imaginable application.
	   //           Probabilities less than 10^-15 are returned as zero.
	   //           However, remember that the formula is only valid for "large" n.
	   // Theta function inversion formula is used for z <= 1
	   //
	   // This function was translated by Rene Brun from PROBKL in CERNLIB.

	   double fj[] = new double[]{-2,-8,-18,-32};
	   double r[] = new double[4];
	   final double w = 2.50662827;
	   // c1 - -pi**2/8, c2 = 9*c1, c3 = 25*c1
	   final double c1 = -1.2337005501361697;
	   final double c2 = -11.103304951225528;
	   final double c3 = -30.842513753404244;

	   double u = Math.abs(z);
	   double p;
	   if (u < 0.2) {
	      p = 1;
	   } else if (u < 0.755) {
	      double v = 1./(u*u);
	      p = 1 - w*(Math.exp(c1*v) + Math.exp(c2*v) + Math.exp(c3*v))/u;
	   } else if (u < 6.8116) {
	      r[1] = 0;
	      r[2] = 0;
	      r[3] = 0;
	      double v = u*u;
	      int maxj = (int)Math.max(1,Math.round(3.0/u));
	      for (int j=0; j<maxj;j++) {
	         r[j] = Math.exp(fj[j]*v);
	      }
	      p = 2*(r[0] - r[1] +r[2] - r[3]);
	   } else {
	      p = 0;
	   }
	   return p;
	   }
	
	public static void main(String args[]){
		double r1[] = new double[1000];
		double r2[] = new double[1000];
		
		for (int i = 0; i < 1000; i++){
			r1[i] = Math.random();
			r2[i] = Math.random();
		}
		Arrays.sort(r1);
		Arrays.sort(r2);
		
		System.out.println("Test result: "+KolmogorovTest(1000, r1, 1000, r2, ""));
	}
}
